#include <windows.h>
#include "hem.h"
#include "strarray.h"
#include "string.h"

/*
struct VS_VERSIONINFO { 
  WORD  wLength; 
  WORD  wValueLength; 
  WORD  wType; 
  WCHAR szKey[]; 
  WORD  Padding1[]; 
  VS_FIXEDFILEINFO Value; 
  WORD  Padding2[]; 
  WORD  Children[]; 
};

struct StringFileInfo {
  WORD        wLength;
  WORD        wValueLength;
  WORD        wType;
  WCHAR       szKey[];
  WORD        Padding[];
  StringTable Children[];
};

struct VarFileInfo {
  WORD  wLength;
  WORD  wValueLength;
  WORD  wType;
  WCHAR szKey[];
  WORD  Padding[];
  Var   Children[];
};
*/

#define SHOWERROR(Msg) {ShowError(Msg); return false;}

/* external definitions */
extern int ShowError(char *Message);

WORD Header[] = L"VS_VERSION_INFO";
WORD Sfi[] = L"StringFileInfo";
WORD Vfi[] = L"VarFileInfo";

/* make these global so that we don't have to pass pointers to them to every function */
LPBYTE lpVsVersionInfo;
DWORD BytesRest;

/* stack related stuff */
#define MAX_PUSH_DEPTH	5
typedef struct _PUSHSTR
{
	LPBYTE lpVsVersionInfo;
	DWORD BytesRest;
} PUSHSTR;

PUSHSTR Stack[MAX_PUSH_DEPTH];
int StackPtr = 0;

//
// throw an exception
//
void Throw()
{
	RaiseException(STATUS_ACCESS_VIOLATION, EXCEPTION_NONCONTINUABLE, 0, 0);
}

//
// Initialize stack
//
void InitStack()
{
	StackPtr = 0;
}

//
// Push/Pop lpVsVersionInfo and BytesRest
//
void Push()
{
	if(StackPtr >= MAX_PUSH_DEPTH) Throw();
	Stack[StackPtr].lpVsVersionInfo = lpVsVersionInfo;
	Stack[StackPtr].BytesRest = BytesRest;
	StackPtr++;
}

void Pop()
{
	if(StackPtr == 0) Throw();
	StackPtr--;
	lpVsVersionInfo = Stack[StackPtr].lpVsVersionInfo;
	BytesRest = Stack[StackPtr].BytesRest;
}

//
// get next byte/word/dword/string with check for data availability
//
void Skip(DWORD N)
{
	if(BytesRest < N) Throw();
	lpVsVersionInfo += N;
	BytesRest -= N;
}

BYTE GetByte()
{
	if(BytesRest < 1) Throw();
	BYTE Result = lpVsVersionInfo[0];
	Skip(1);
	return Result;
}

WORD GetWord()
{
	if(BytesRest < 2) Throw();
	WORD Result = (lpVsVersionInfo[1] << 8) | lpVsVersionInfo[0];
	Skip(2);
	return Result;
}

DWORD GetDword()
{
	if(BytesRest < 4) Throw();
	DWORD Result = (lpVsVersionInfo[3] << 24) | (lpVsVersionInfo[2] << 16) | (lpVsVersionInfo[1] << 8) | lpVsVersionInfo[0];
	Skip(4);
	return Result;
}

DWORD GetString(PWORD OutString, DWORD MaxLength)
{
	DWORD i = 0;
	WORD Ch;
	do
	{
		Ch = GetWord();
		OutString[i++] = Ch;
		if(i >= MaxLength) Throw();
	}
	while(Ch != 0);
	return i;
}

void GetStringN(PWORD OutString, DWORD Len)
{
	for(DWORD i = 0; i < Len; i++)
	{
		WORD Ch = GetWord();
		OutString[i] = Ch;
		/* the cure for really sick cases like X9.SYS & X9TT.SYS from Compuware DriverStudio that have the Len field totally unrelated to reality */
		if(Ch == 0) break;
	}
	OutString[i] = 0;
}

//
// Switch to nested structure which starts with word-wide length field
//
void NestedStructure()
{
	/* check that current state is pushed so that we can return to it later */
	if(StackPtr == 0) Throw();
	BytesRest = GetWord();
	lpVsVersionInfo -= 2;
}

//
// Append formatted value
//
void AppendFormat(char *Format, int Value)
{
	char Buffer[1024];
	wsprintf(Buffer, Format, Value);
	StringAppend(Buffer);
}

void AppendFormat2(char *Format, LPVOID Value1, DWORD Value2)
{
	char Buffer[1024];
	wsprintf(Buffer, Format, Value1, Value2);
	StringAppend(Buffer);
}

//
// Parse FILEFLAGS flags
//
void ParseFileFlags(DWORD Flags)
{
	StringEmpty();
	if(Flags)
	{
		if(Flags & ~VS_FFI_FILEFLAGSMASK) AppendFormat("0x%8x | ", Flags & ~VS_FFI_FILEFLAGSMASK);
		if((Flags & VS_FFI_FILEFLAGSMASK) == VS_FFI_FILEFLAGSMASK) StringAppend("VS_FFI_FILEFLAGSMASK | ");
		else
		{
			if(Flags & VS_FF_SPECIALBUILD) StringAppend("VS_FF_SPECIALBUILD | ");
			if(Flags & VS_FF_INFOINFERRED) StringAppend("VS_FF_INFOINFERRED | ");
			if(Flags & VS_FF_PRIVATEBUILD) StringAppend("VS_FF_PRIVATEBUILD | ");
			if(Flags & VS_FF_PATCHED)      StringAppend("VS_FF_PATCHED | ");
			if(Flags & VS_FF_PRERELEASE)   StringAppend("VS_FF_PRERELEASE | ");
			if(Flags & VS_FF_DEBUG)        StringAppend("VS_FF_DEBUG | ");
		}
		StringMakeShorter(sizeof(" | ") - 1);
	}
}

//
// Process FILEFLAGSMASK field
//
void FileFlagsMask(DWORD Ffm)
{
	if(Ffm)
	{
		char Buffer[1024];
		ParseFileFlags(Ffm);
		wsprintf(Buffer, " FILEFLAGSMASK  %s /* 0x%xL */", StringGetString(), Ffm);
		StrArrayAdd(Buffer);
	}
	else StrArrayAdd(" FILEFLAGSMASK  0x0L");
}

//
// Process FILEFLAGS field
//
void FileFlags(DWORD Ff)
{
	if(Ff)
	{
		char Buffer[1024];
		ParseFileFlags(Ff);
		wsprintf(Buffer, " FILEFLAGS      %s /* 0x%xL */", StringGetString(), Ff);
		StrArrayAdd(Buffer);
	}
	else StrArrayAdd(" FILEFLAGS      0x0L");
}

//
// Process FILEOS field
//
void FileOS(DWORD Os)
{
	char Buffer[1024];
	StringEmpty();
	switch(Os)
	{
		case VOS_DOS_WINDOWS16:		StringAppend("VOS_DOS_WINDOWS16");	break;
		case VOS_DOS_WINDOWS32:		StringAppend("VOS_DOS_WINDOWS32");	break;
		case VOS_OS216_PM16:		StringAppend("VOS_OS216_PM16");		break;
		case VOS_OS232_PM32:		StringAppend("VOS_OS232_PM32");		break;
		case VOS_NT_WINDOWS32:		StringAppend("VOS_NT_WINDOWS32");	break;
		default:
			switch(Os & 0xffff0000)
			{
				case VOS_UNKNOWN:			StringAppend("VOS_UNKNOWN");			break;
				case VOS_DOS:				StringAppend("VOS_DOS");				break;
				case VOS_OS216:				StringAppend("VOS_OS216");				break;
				case VOS_OS232:				StringAppend("VOS_OS232");				break;
				case VOS_NT:				StringAppend("VOS_NT");					break;
				default:					AppendFormat("0x%04x", HIWORD(Os));
			}
			StringAppend(" | ");
			switch(Os & 0xffff)
			{
				case VOS__BASE:				StringAppend("VOS__BASE");				break;
				case VOS__WINDOWS16:		StringAppend("VOS__WINDOWS16");			break;
				case VOS__PM16:				StringAppend("VOS__PM16");				break;
				case VOS__PM32:				StringAppend("VOS__PM32");				break;
				case VOS__WINDOWS32:		StringAppend("VOS__WINDOWS32");			break;
				default:					AppendFormat("0x%04x", LOWORD(Os));
			}
	}
	wsprintf(Buffer, " FILEOS         %s /* 0x%xL */", StringGetString(), Os);
	StrArrayAdd(Buffer);
}

//
// Process FILETYPE/FILESUBTYPE fields
//
char VFTFormat[] = "%s /* 0x%xL */";

void FFTDrv(DWORD SubType)
{
	switch(SubType)
	{
		case VFT2_UNKNOWN:			AppendFormat2(VFTFormat, "VFT2_UNKNOWN",         VFT2_UNKNOWN);			break;
		case VFT2_DRV_PRINTER:		AppendFormat2(VFTFormat, "VFT2_DRV_PRINTER",     VFT2_DRV_PRINTER);		break;
		case VFT2_DRV_KEYBOARD:		AppendFormat2(VFTFormat, "VFT2_DRV_KEYBOARD",    VFT2_DRV_KEYBOARD);	break;
		case VFT2_DRV_LANGUAGE:		AppendFormat2(VFTFormat, "VFT2_DRV_LANGUAGE",    VFT2_DRV_LANGUAGE);	break;
		case VFT2_DRV_DISPLAY:		AppendFormat2(VFTFormat, "VFT2_DRV_DISPLAY",     VFT2_DRV_DISPLAY);		break;
		case VFT2_DRV_MOUSE:		AppendFormat2(VFTFormat, "VFT2_DRV_MOUSE",       VFT2_DRV_MOUSE);		break;
		case VFT2_DRV_NETWORK:		AppendFormat2(VFTFormat, "VFT2_DRV_NETWORK",     VFT2_DRV_NETWORK);		break;
		case VFT2_DRV_SYSTEM:		AppendFormat2(VFTFormat, "VFT2_DRV_SYSTEM",      VFT2_DRV_SYSTEM);		break;
		case VFT2_DRV_INSTALLABLE:	AppendFormat2(VFTFormat, "VFT2_DRV_INSTALLABLE", VFT2_DRV_INSTALLABLE);	break;
		case VFT2_DRV_SOUND:		AppendFormat2(VFTFormat, "VFT2_DRV_SOUND",       VFT2_DRV_SOUND);		break;
		case VFT2_DRV_COMM:			AppendFormat2(VFTFormat, "VFT2_DRV_COMM",        VFT2_DRV_COMM);		break;
		case VFT2_DRV_INPUTMETHOD:	AppendFormat2(VFTFormat, "VFT2_DRV_INPUTMETHOD", VFT2_DRV_INPUTMETHOD);	break;
		default:					AppendFormat("0x%xL", SubType);
	}
}

void FFTFont(DWORD SubType)
{
	switch(SubType)
	{
		case VFT2_FONT_RASTER:		AppendFormat2(VFTFormat, "VFT2_FONT_RASTER",   VFT2_FONT_RASTER);		break;
		case VFT2_FONT_VECTOR:		AppendFormat2(VFTFormat, "VFT2_FONT_VECTOR",   VFT2_FONT_VECTOR);		break;
		case VFT2_FONT_TRUETYPE:	AppendFormat2(VFTFormat, "VFT2_FONT_TRUETYPE", VFT2_FONT_TRUETYPE);		break;
		default:					AppendFormat("0x%xL", SubType);
	}
}

void FFTVxD(DWORD SubType)
{
	AppendFormat("0x%xL (Virtual Device ID)", SubType);
}

void FileType(DWORD Type, DWORD SubType)
{
	StringEmpty();
	StringAppend(" FILETYPE       ");
	switch(Type)
	{
		case VFT_UNKNOWN:			AppendFormat2(VFTFormat, "VFT_UNKNOWN",    VFT_UNKNOWN);	break;
		case VFT_APP:				AppendFormat2(VFTFormat, "VFT_APP",        VFT_APP);		break;
		case VFT_DLL:				AppendFormat2(VFTFormat, "VFT_DLL",        VFT_DLL);		break;
		case VFT_DRV:				AppendFormat2(VFTFormat, "VFT_DRV",        VFT_DRV);		break;
		case VFT_FONT:				AppendFormat2(VFTFormat, "VFT_FONT",       VFT_FONT);		break;
		case VFT_VXD:				AppendFormat2(VFTFormat, "VFT_VXD",        VFT_VXD);		break;
		case VFT_STATIC_LIB:		AppendFormat2(VFTFormat, "VFT_STATIC_LIB", VFT_STATIC_LIB);	break;
		default:					AppendFormat("0x%xL", Type);
	}
	StrArrayAdd(StringGetString());

	StringEmpty();
	StringAppend(" FILESUBTYPE    ");
	switch(Type)
	{
		case VFT_DRV:				FFTDrv(SubType);	break;
		case VFT_FONT:				FFTFont(SubType);	break;
		case VFT_VXD:				FFTVxD(SubType);	break;
		default:					AppendFormat("0x%xL", SubType);
	}
	StrArrayAdd(StringGetString());
}

//
// process the VS_FIXEDFILEINFO structure
//
bool ProcessFFI()
{
	char Buffer[2048];
	
	VS_FIXEDFILEINFO *lpFFI = (VS_FIXEDFILEINFO *)(lpVsVersionInfo);
	if(lpFFI->dwSignature == VS_FFI_SIGNATURE && lpFFI->dwStrucVersion == VS_FFI_STRUCVERSION)
	{
		wsprintf(Buffer, " FILEVERSION    %d, %d, %d, %d", HIWORD(lpFFI->dwFileVersionMS), LOWORD(lpFFI->dwFileVersionMS), HIWORD(lpFFI->dwFileVersionLS), LOWORD(lpFFI->dwFileVersionLS));
		StrArrayAdd(Buffer);

		wsprintf(Buffer, " PRODUCTVERSION %d, %d, %d, %d", HIWORD(lpFFI->dwProductVersionMS), LOWORD(lpFFI->dwProductVersionMS), HIWORD(lpFFI->dwProductVersionLS), LOWORD(lpFFI->dwProductVersionLS));
		StrArrayAdd(Buffer);

		FileFlagsMask(lpFFI->dwFileFlagsMask);
		FileFlags(lpFFI->dwFileFlags);
		FileOS(lpFFI->dwFileOS);
		FileType(lpFFI->dwFileType, lpFFI->dwFileSubtype);

		Skip(sizeof(VS_FIXEDFILEINFO));
		return true;
	}
	SHOWERROR("Invalid VS_FIXEDINFO structure");
}

//
// process padding to DWORD
//
void Pad32()
{
	DWORD PaddingBytes = (((DWORD)lpVsVersionInfo + 3) & -4) - (DWORD)lpVsVersionInfo;
	Skip(PaddingBytes);
}

//
// Convert hex string to number. No checks for validity here
//
DWORD HexToInt(PCHAR Buffer, int Digits)
{
	DWORD Val = 0;
	for(int i = 0; i < Digits; i++)
	{
		char Ch = Buffer[i];
		Ch = toupper(Ch);
		Val <<= 4;
		if(Ch >= 'A') Val += Ch - 'A' + 10;
		else Val += Ch - '0';
	}
	return Val;
}

//
// Process array of String structures which is part of StringFileInfo.Children
//
bool ProcessStrings()
{
	Pad32();	/* this is important for empty arrays */
	while(BytesRest)
	{
		Pad32();
		if(BytesRest == 0) break;

		Push();
		NestedStructure();

		DWORD StringLen = GetWord();
		DWORD ValueLen = GetWord();
		DWORD wType = GetWord();
		if(wType != 1 && wType != 0) SHOWERROR("Unknown String.wType value");

		WORD Key[512], Value[512];
		char KeyAnsi[512], ValueAnsi[512], Buffer[2048];
		GetString(Key, sizeof(Key));
		Pad32();
		if(ValueLen >= sizeof(Value)) SHOWERROR("Too long string");
		
		/* a special case: in contrast to documentation ValueLen may be given in bytes, not in words. See %WINNT%/Setup1.exe */
		if((ValueLen * 2) > BytesRest) ValueLen >>= 1;

		GetStringN(Value, ValueLen);

		WideCharToMultiByte(CP_OEMCP, 0, Key, -1, KeyAnsi, sizeof(KeyAnsi), 0, 0);
		WideCharToMultiByte(CP_OEMCP, 0, Value, -1, ValueAnsi, sizeof(ValueAnsi), 0, 0);
		wsprintf(Buffer, "   VALUE \"%s\", \"%s\"", KeyAnsi, ValueAnsi);
		StrArrayAdd(Buffer);

		Pop();
		Skip(StringLen);
	}
	return true;
}

//
// Process StringFileInfo structure
//
bool ProcessStringFileInfo()
{
	Pad32();	/* this is important for empty arrays */
	while(BytesRest)
	{
		Pad32();
		if(BytesRest == 0) break;

		Push();
		NestedStructure();

		DWORD ChildLen = GetWord();
		if(GetWord()) SHOWERROR("StringTable.wValueLength != 0");
		WORD wType = GetWord();
		if(wType != 1 && wType != 0) SHOWERROR("Unknown StringTable.wType value");

		char LangBuffer[16], Buffer[512], LangName[256];
		for(int i = 0; i < 8; i++)
		{
			WORD Ch = GetWord();
			if(HIBYTE(Ch) || !isxdigit(LOBYTE(Ch))) SHOWERROR("Invalid language ID in StringTable.szKey");
			LangBuffer[i] = LOBYTE(Ch);
		}
		LangBuffer[8] = 0;

		DWORD LangId = HexToInt(LangBuffer, 4);
		DWORD CodePage = HexToInt(&LangBuffer[4], 4);

		if(!VerLanguageName(LangId, LangName, sizeof(LangName))) strcpy(LangName, "unknown");

		wsprintf(Buffer, "  BLOCK \"%s\" /* %s, %d */", LangBuffer, LangName, CodePage);
		StrArrayAdd(Buffer);

		StrArrayAdd("  {");
		if(!ProcessStrings()) return false;
		StrArrayAdd("  }");

		Pop();
		Skip(ChildLen);
	}
	return true;
}

//
// Process VarFileInfo structure. Note that Borland guys ignore the docs (just a habit,
// I suppose...) and allow an array of Var structures here
//
bool ProcessVarFileInfo()
{
	while(BytesRest)
	{
		Push();
		NestedStructure();
		
		DWORD VarLen = GetWord();
		DWORD ValLen = GetWord();
		WORD wType = GetWord();
		if(wType != 1 && wType != 0) SHOWERROR("Unknown Var.wType value");
		
		WORD Key[512];
		char KeyAnsi[512], Buffer[1024];
		GetString(Key, sizeof(Key));
		Pad32();
		
		WideCharToMultiByte(CP_OEMCP, 0, Key, -1, KeyAnsi, sizeof(KeyAnsi), 0, 0);
		wsprintf(Buffer, "  VALUE \"%s\"", KeyAnsi);
		StringEmpty();
		StringAppend(Buffer);
		
		DWORD NumPairs = ValLen >> 2;
		for(DWORD i = 0; i < NumPairs; i++)
		{
			DWORD Lang = GetWord();
			DWORD Cp = GetWord();

			char LangName[256], LangBuffer[512];
			if(!VerLanguageName(Lang, LangName, sizeof(LangName))) strcpy(LangName, "unknown");
			wsprintf(LangBuffer, "0x%x /* %s */, ", Lang, LangName);
			AppendFormat2(", %s%d", LangBuffer, Cp);
		}
		
		StrArrayAdd(StringGetString());
		
		Pop();
		Skip(VarLen);
	}
	return true;
}

//
// process the Children field of the VS_VERSIONINFO structure which is an array of StringFileInfo and VarFileInfo structures
//
bool ProcessChildren()
{
	StrArrayAdd("{");

	/* There may be any amount (including zero) of nested structures. Let's process them all while there are bytes to process */
	Pad32();	/* this is important for empty arrays */
	while(BytesRest)
	{
		Pad32();
		if(BytesRest == 0) break;

		Push();
		NestedStructure();

		DWORD ChildLen = GetWord();
		/* size of (String|Var)FileInfo structures canot be less than 8 bytes */
		if(ChildLen < 8) SHOWERROR("Invalid [String|Var]FileInfo structure");
		if(GetWord()) SHOWERROR("[String|Var]FileInfo.wValueLength != 0");
		WORD wType = GetWord();
		/* MS and Borland guys fill the following field differently */
		if(wType != 1 && wType != 0) SHOWERROR("Unknown [String|Var]FileInfo.wType value");
		
		WORD Buffer[256];

		GetString(Buffer, sizeof(Buffer));
		Pad32();

		if(wcscmp(Buffer, Sfi) == 0)
		{
			StrArrayAdd(" BLOCK \"StringFileInfo\"");
			StrArrayAdd(" {");
			if(!ProcessStringFileInfo()) return false;
			StrArrayAdd(" }");
		}
		else
			if(wcscmp(Buffer, Vfi) == 0)
			{
				StrArrayAdd(" BLOCK \"VarFileInfo\"");
				StrArrayAdd(" {");
				if(!ProcessVarFileInfo()) return false;
				StrArrayAdd(" }");
			}
			else SHOWERROR("Unknown child type");
		
		Pop();
		Skip(ChildLen);
	}
	StrArrayAdd("}");
	return true;
}

//
// process the VS_VERSIONINFO structure
//
bool ProcessVersionInfo(PBYTE lpRes, DWORD ResSize)
{
	lpVsVersionInfo = lpRes;
	BytesRest = ResSize;
	InitStack();

	__try
	{
		
		/* check the size of VS_VERSIONINFO */
		DWORD Size = GetWord();
		if(Size == ResSize)
		{
			DWORD ValueLength = GetWord();
			if(ValueLength != sizeof(VS_FIXEDFILEINFO) && ValueLength != 0) SHOWERROR("VS_VERSIONINFO.ValueLength field is invalid");

			DWORD Type = GetWord();
			if(Type != 0) SHOWERROR("VS_VERSIONINFO.wType != 0");

			if(BytesRest > sizeof(Header))
			{
				/* check for unicode "VS_VERSION_INFO" string */
				if(memcmp(lpVsVersionInfo, Header, sizeof(Header)) == 0)
				{
					Skip(sizeof(Header) + 2);
					StrArrayAdd("VS_VERSION_INFO VERSIONINFO");

					/* a VS_FIXEDFILEINFO structure follows immediately if ValueLength is not 0 */
					if(ValueLength)
					{
						if(BytesRest < sizeof(VS_FIXEDFILEINFO)) SHOWERROR("Invalid VS_FIXEDINFO structure");
						if(!ProcessFFI()) return false;
						return ProcessChildren();
					}
					return true;
				}
			}
		}
		ShowError("Invalid VS_VERSIONINFO structure");
	}
	__except(EXCEPTION_EXECUTE_HANDLER)
	{
		ShowError("Compiled VS_VERSIONINFO structure is invalid");
	}
	return false;
}
